@@ -51,8 +51,8 @@
 //C- | MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 //C- +------------------------------------------------------------------
 // 
-// $Id: DataPool.cpp,v 1.10 2003/11/07 22:08:20 leonb Exp $
-// $Name: release_3_5_14 $
+// $Id: DataPool.cpp,v 1.11 2004/08/06 15:11:29 leonb Exp $
+// $Name: release_3_5_15 $
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -68,13 +68,9 @@
 #include "GURL.h"
 #include "debug.h"
 
-#if defined(macintosh) //MCW can't compile
-#else
-#ifndef UNDER_CE 
-#include <sys/types.h>
+#ifndef macintosh
+# include <sys/types.h>
 #endif
-#endif
-
 
 #ifdef HAVE_NAMESPACES
 namespace DJVU {
