@@ -52,7 +52,7 @@
 //C- +------------------------------------------------------------------
 //C- */ 
 
-/* $Id: ddjvuapi.h,v 1.15 2004/08/03 23:43:22 leonb Exp $ */
+/* $Id: ddjvuapi.h,v 1.29 2005/06/08 12:07:39 leonb Exp $ */
 
 #ifndef DDJVUAPI_H
 #define DDJVUAPI_H
@@ -63,6 +63,10 @@ extern "C" {
 #if 0
 }
 #endif
+
+#include <stdlib.h>
+#include <stdio.h>
+
 #ifndef DDJVUAPI
 # define DDJVUAPI /**/
 #endif
@@ -73,6 +77,8 @@ extern "C" {
 # define FALSE (0)
 #endif
 
+
+
 /* -------------------------------------------------- */
 /* DDJVU API                                          */
 /* -------------------------------------------------- */
@@ -88,6 +94,8 @@ extern "C" {
    complexity with a familiar event model.
 */
 
+#define DDJVUAPI_VERSION 15
+
 typedef struct ddjvu_context_s  ddjvu_context_t;
 typedef union  ddjvu_message_s  ddjvu_message_t;
 typedef struct ddjvu_job_s      ddjvu_job_t;
@@ -120,7 +128,6 @@ typedef struct ddjvu_format_s   ddjvu_format_t;
      <"doc/djvu3changes.txt"> and <"doc/djvu2spec.djvu">.
 */
 
-
   
 
 /* -------------------------------------------------- */
@@ -181,6 +188,31 @@ ddjvu_cache_clear(ddjvu_context_t *context);
 
 /* ------- MESSAGE QUEUE ------- */
 
+/* Messages produced by the ddjvu api accumulate into
+   the message queue. Processing messages is not optional!
+   A typical message handling routine looks like this:
+   
+   void handle_ddjvu_messages(ddjvu_context_t *ctx, int wait)
+   {
+     const ddjvu_message_t *msg;
+     if (wait)
+       ddjvu_message_wait(ctx);
+     while ((msg = ddjvu_message_peek(ctx)))
+     {
+       switch(msg->m_any.tag)
+       { 
+       case DDJVU_ERROR:      .... ; break;
+       case DDJVU_INFO:       .... ; break;
+       case DDJVU_NEWSTREAM:  .... ; break;
+       ....
+       default: break;
+       }
+       ddjvu_message_pop(ctx);
+     }
+   }
+*/
+
+
 /* ddjvu_message_peek ---
    Returns a pointer to the next DDJVU message.
    This function returns 0 if no message is available.
@@ -215,7 +247,8 @@ ddjvu_message_pop(ddjvu_context_t *context);
    a new message is posted to the ddjvuapi message queue,
    and returns a pointer to the previous callback function.
    This callback function can be called at any time
-   while other code is executing. It should simply signal
+   while other code is executing. Do not call ddjvuapi
+   functions from this callback. It should simply signal
    the main application event loop that new ddjvuapi messages
    are available.  Under WIN32, this is usually achieved
    by posting a user window message.  Under UNIX, this is
@@ -278,8 +311,9 @@ ddjvu_job_stop(ddjvu_job_t *job);
 /* ddjvu_job_set_user_data ---
    ddjvu_job_get_user_data ---
    Each job can store an arbitray pointer
-   that callers can use for any purpose. These two functions
-   provide for accessing or setting this pointer. */
+   that callers can use for any purpose. These two 
+   functions provide for accessing or setting this pointer. 
+   This pointer is cleared when the job is released */
 
 DDJVUAPI void
 ddjvu_job_set_user_data(ddjvu_job_t *job, void *userdata);
@@ -289,11 +323,11 @@ ddjvu_job_get_user_data(ddjvu_job_t *job);
 
 
 /* ddjvu_job_release ---
-   Releases a reference to a job object.
-   This will not cause the job to stop executing.
+   Releases a reference to a job object and clears its user 
+   data field.  This does not cause the job to stop executing.
    The calling program should no longer reference this object.
-   The object itself will be destroyed as soon as no other object
-   or thread needs it. */
+   The object itself will be destroyed as soon as no 
+   other object or thread needs it. */
 
 DDJVUAPI void
 ddjvu_job_release(ddjvu_job_t *job);
@@ -393,7 +427,7 @@ struct ddjvu_message_info_s {   /* ddjvu_message_t::m_info */
 
    Argument <url> specifies an optional URL for the document.  
    The URL follows the usual syntax (<"protocol://machine/path">). 
-   It only serves two purposes:
+   It should not end with a slash. It only serves two purposes:
    - The URL is used as a key for the cache of decoded pages.
    - The URL is used to document <m_newstream> messages.
 
@@ -578,6 +612,36 @@ DDJVUAPI int
 ddjvu_document_get_pagenum(ddjvu_document_t *document);
 
 
+/* ddjvu_document_get_pageinfo ---
+   Attempts to obtain information about page <pageno>
+   without decoding the page. If the information is available,
+   the function returns <DDJVU_JOB_OK> and fills the <info> structure. 
+   Otherwise it starts fetching page data and returns <DDJVU_JOB_STARTED>. 
+   This function causes the emission of <m_pageinfo> messages 
+   with zero in the <m_any.page> field.
+   Typical synchronous usage:
+
+   ddjvu_status_t r;
+   ddjvu_pageinfo_t info;
+   while ((r=ddjvu_document_get_pageinfo(doc,pageno,&info))<DDJVU_JOB_OK)
+     handle_ddjvu_messages(ctx, TRUE);
+   if (r>=DDJVU_JOB_FAILED)
+     signal_error();
+
+   When the djvu document comes from the network, the above idiom 
+   is very slow because it waits until the data for all page is present. 
+*/      
+
+typedef struct ddjvu_pageinfo_s {
+  int width;
+  int height;
+  int dpi;
+} ddjvu_pageinfo_t;
+
+DDJVUAPI ddjvu_status_t
+ddjvu_document_get_pageinfo(ddjvu_document_t *document, int pageno, 
+                            ddjvu_pageinfo_t *info);
+
 
 
 /* -------------------------------------------------- */
@@ -657,8 +721,16 @@ ddjvu_page_job(ddjvu_page_t *page);
 
 
 /* ddjvu_message_t::m_pageinfo ---
-   This message is generated when the basic 
-   information about a page is available. */
+   The page decoding process generates this message
+   - when basic page information is available and 
+     before any <m_relayout> or <m_redisplay> message,
+   - when the page decoding thread terminates.
+   You can distinguish both cases using 
+   function ddjvu_page_decoding_done().
+   Messages <m_pageinfo> are also generated as a consequence of 
+   functions such as <ddjvu_document_get_pageinfo>. 
+   The field <m_any.page> of such message is null.
+*/
 
 struct ddjvu_message_pageinfo_s {  /* ddjvu_message_t::m_pageinfo */
   ddjvu_message_any_t  any;
@@ -697,6 +769,17 @@ struct ddjvu_message_chunk_s {     /* ddjvu_message_t::m_chunk */
   const char *chunkid;
 }; 
 
+/* About page messages --
+   Both the <m_relayout> and <m_redisplay> messages are derived from the
+   <m_chunk> message.  They are intended for driving a djvu image viewer. 
+   When receiving <m_relayout>, the viewer should get the image size, decide
+   zoom factors, and place the image area, scrollbars, toolbars, and other gui
+   objects.  When receiving <m_redisplay>, the viewer should invalidate the
+   image area so that the gui toolkint calls the repaint event handler. This
+   handler should call ddjvu_page_render() and paint the part og the
+   image that needs repainting. */
+
+
 
 /* ------- QUERIES ------- */
 
@@ -806,11 +889,11 @@ ddjvu_page_set_rotation(ddjvu_page_t *page,
 
 /* ddjvu_rect_t ---
    This structure specifies the location of a rectangle.
-   Coordinates are always expressed in pixels relative to
-   the BOTTOM LEFT CORNER of an image.  Members <x> and <y>
-   indicate the position of the bottom left corner of the
-   rectangle Members <w> and <h> indicate the width and
-   height of the rectangle. */
+   Coordinates are usually expressed in pixels relative to 
+   the BOTTOM LEFT CORNER (but see ddjvu_format_set_y_direction).
+   Members <x> and <y> indicate the position of the bottom left 
+   corner of the rectangle Members <w> and <h> indicate the 
+   width and height of the rectangle. */
 
 typedef struct ddjvu_rect_s {
   int x, y;
@@ -905,7 +988,7 @@ ddjvu_format_create(ddjvu_format_style_t style,
                     int nargs, unsigned int *args);
 
 
-/* ddjvu_format_set_top_to_bottom ---
+/* ddjvu_format_set_row_order ---
    Sets a flag indicating whether the rows in the pixel buffer
    are stored starting from the top or the bottom of the image.
    Default ordering starts from the bottom of the image.
@@ -915,6 +998,17 @@ DDJVUAPI void
 ddjvu_format_set_row_order(ddjvu_format_t *format, int top_to_bottom);
 
 
+/* ddjvu_format_set_y_direction ---
+   Sets a flag indicating whether the y coordinates in the drawing 
+   area are oriented from bottom to top, or from top to botttom.  
+   The default is bottom to top, similar to PostScript.
+   This is the opposite of the X11 convention. */
+
+DDJVUAPI void
+ddjvu_format_set_y_direction(ddjvu_format_t *format, int top_to_bottom);
+
+
+
 /* ddjvu_format_set_ditherbits ---
    Specifies the final depth of the image on the screen.
    This is used to decide which dithering algorithm should be used.
@@ -992,14 +1086,11 @@ ddjvu_thumbnail_render(ddjvu_document_t *document, int pagenum,
 /* SAVE AND PRINT JOBS                                */
 /* -------------------------------------------------- */
 
-#ifndef DDJVU_WITHOUT_STDIO
-# include <stdlib.h>
-# include <stdio.h>
-#else 
-# define FILE void
-#endif
 
-/* Not yet implemented */
+
+/* ddjvu_message_t::m_progress ---
+   These messages are generated to indicate progress 
+   towards the completion of a print or save job. */
 
 struct ddjvu_message_progress_s {
   ddjvu_message_any_t any;
@@ -1007,31 +1098,72 @@ struct ddjvu_message_progress_s {
   int percent;
 };
 
+/* ddjvu_document_print ---
+   Converts specified pages of a djvu document into postscript.  
+   This function works asynchronously in a separate thread.
+   You can use the following idiom for synchronous operation:
 
-/* Options like ddjvu ? */
-DDJVUAPI ddjvu_job_t *
-ddjvu_document_export(ddjvu_document_t *document, FILE *output,
-                      int optc, const char * const * optv);
+     ddjvu_job_t *job = ddjvu_document_print(....);
+     while (! ddjvu_job_done(job) )
+       handle_ddjvu_messages(context, TRUE);
+       
+   The postscript data is written to stdio file <output>.
+   Arguments <optc> and <optv> specify printing options.
+   All options described on the <djvups> man page are 
+   recognized, except <"-help"> and <"-verbose">.
+*/
 
-
-/* Options like djvups ? */
 DDJVUAPI ddjvu_job_t *
 ddjvu_document_print(ddjvu_document_t *document, FILE *output,
                      int optc, const char * const * optv);
 
 
-/* Options to be defined */
+/* Not yet implemented ---
+   Saves the djvu document as a bundled djvu file.
+   This function works asynchronously in a separate thread.
+   You can use the following idiom for synchronous operation:
+
+     ddjvu_job_t *job = ddjvu_document_save(....);
+     while (! ddjvu_job_done(job) )
+       handle_ddjvu_messages(context, TRUE);
+     
+   The bundled djvu data is written to file <output>
+   which must be seekable. Arguments <optc> and <optv> 
+   exactly like command line arguments of a program.
+   The only supported option is "-page=<pagespec>".
+   See the man page for <djvups> for more information
+   about page specifications.
+*/
 DDJVUAPI ddjvu_job_t *
 ddjvu_document_save(ddjvu_document_t *document, FILE *output, 
                     int optc, const char * const * optv);
 
 
 
+/* -------------------------------------------------- */
+/* S-EXPRESSIONS                                      */
+/* -------------------------------------------------- */
+
+/* Not yet defined */
+
+/* Not yet implemented */
+
+
 /* -------------------------------------------------- */
 /* ANNOTATIONS AND HIDDEN_TEXT                        */
 /* -------------------------------------------------- */
 
 /* Not yet defined */
+
+/* Not yet implemented */
+
+
+/* -------------------------------------------------- */
+/* HIDDEN_TEXT                                        */
+/* -------------------------------------------------- */
+
+/* Not yet defined */
+
 /* Not yet implemented */
 
 
@@ -1053,10 +1185,10 @@ union ddjvu_message_s {
   struct ddjvu_message_relayout_s   m_relayout;
   struct ddjvu_message_redisplay_s  m_redisplay;
   struct ddjvu_message_thumbnail_s  m_thumbnail;
+  struct ddjvu_message_progress_s   m_progress;
 };
 
 
-
 /* -------------------------------------------------- */
 /* BACKDOORS                                          */
 /* -------------------------------------------------- */
@@ -1070,8 +1202,8 @@ union ddjvu_message_s {
    These functions provide an access to the libdjvu objects 
    associated with the ddjvuapi objects.  These backdoors can
    be useful for advanced manipulations.  These two functions 
-   are declared in C++ when file "ddjvuapi.h" is included 
-   after the libdjvu header files "DjVuDocument.h". */
+   are declared in C++ when file <"ddjvuapi.h"> is included 
+   after the libdjvu header files <"DjVuDocument.h">. */
 
 #ifdef __cplusplus
 # ifndef NOT_USING_DJVU_NAMESPACE
