@@ -51,7 +51,7 @@
 //C- | MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 //C- +------------------------------------------------------------------
 // 
-// $Id: ddjvu.cpp,v 1.19 2004/07/24 16:16:17 leonb Exp $
+// $Id: ddjvu.cpp,v 1.23 2005/06/08 12:02:27 leonb Exp $
 
 
 /* Program ddjvu has been rewritten to use the ddjvuapi only.
@@ -67,8 +67,13 @@
 #include <stdarg.h>
 #include <string.h>
 #include <locale.h>
+#include <fcntl.h>
 #include <errno.h>
 
+#ifdef UNIX
+# include <sys/time.h>
+#endif
+
 #include "libdjvu/ddjvuapi.h"
 
 #if defined(WIN32) || defined(__CYGWIN32__)
@@ -97,9 +102,23 @@
 # define I18N(x) (x)
 #endif
 
+
+unsigned long 
+ticks(void)
+{
+#ifdef UNIX
+  struct timeval tv;
+  if (gettimeofday(&tv, NULL) >= 0)
+    return (unsigned long)(((tv.tv_sec & 0xfffff)*1000)+(tv.tv_usec/1000));
+#endif
+  return 0;
+}
+
 ddjvu_context_t *ctx;
 ddjvu_document_t *doc;
 
+unsigned long timingdata[4];
+
 double       flag_scale = -1;
 int          flag_size = -1;
 int          flag_aspect = -1;
@@ -121,6 +140,9 @@ TIFF *tiff;
 #endif
 
 
+
+/* Djvuapi events */
+
 void
 handle(int wait)
 {
@@ -139,9 +161,6 @@ handle(int wait)
             fprintf(stderr,"ddjvu: '%s:%d'\n", 
                     msg->m_error.filename, msg->m_error.lineno);
           exit(10);
-        case DDJVU_INFO:
-          fprintf(stderr,"ddjvu info: %s\n", msg->m_info.message);
-          break;
         default:
           break;
         }
@@ -150,7 +169,6 @@ handle(int wait)
 }
 
 
-
 void 
 die(const char *fmt, ...)
 {
@@ -167,6 +185,7 @@ die(const char *fmt, ...)
   exit(10);
 }
 
+
 void
 inform(ddjvu_page_t *page, int pageno)
 {
@@ -189,11 +208,13 @@ inform(ddjvu_page_t *page, int pageno)
         fprintf(stderr,"%s\n", description);
       if (description)
         free(description);
+      if (timingdata[0] != timingdata[1])
+	fprintf(stderr,"Decoding time:  %5ld ms\n",
+		timingdata[1] - timingdata[0] );
     }
 }
 
 
-
 void
 render(ddjvu_page_t *page)
 {
@@ -330,8 +351,14 @@ render(ddjvu_page_t *page)
     die(i18n("Cannot allocate image buffer"));
 
   /* Render */
+  timingdata[2] = ticks();
   if (! ddjvu_page_render(page, mode, &prect, &rrect, fmt, rowsize, image))
     die(i18n("Cannot render image"));
+  timingdata[3] = ticks();
+  if (flag_verbose)
+    if (timingdata[2] != timingdata[3])
+      fprintf(stderr,"Rendering time: %5ld ms\n",
+	      timingdata[3] - timingdata[2] );
 
   /* Output */
   switch (flag_format)
@@ -407,8 +434,10 @@ render(ddjvu_page_t *page)
         char *s = image;
         TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, (uint32)rrect.w);
         TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, (uint32)rrect.h);
-        TIFFSetField(tiff, TIFFTAG_XRESOLUTION, (float)((dpi*rrect.w+iw/2)/iw));
-        TIFFSetField(tiff, TIFFTAG_YRESOLUTION, (float)((dpi*rrect.h+ih/2)/ih));
+        TIFFSetField(tiff, TIFFTAG_XRESOLUTION, 
+		     (float)((dpi*rrect.w+iw/2)/iw));
+        TIFFSetField(tiff, TIFFTAG_YRESOLUTION, 
+		     (float)((dpi*rrect.h+ih/2)/ih));
         TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
         TIFFSetField(tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
         TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, (uint32)64);
@@ -467,10 +496,12 @@ dopage(int pageno)
 {
   ddjvu_page_t *page;
   /* Decode page */
+  timingdata[0] = ticks();
   if (! (page = ddjvu_page_create_by_pageno(doc, pageno-1)))
     die(i18n("Cannot access page %d."), pageno);
   while (! ddjvu_page_decoding_done(page))
     handle(TRUE);
+  timingdata[1] = ticks();
   /* Open files */
   if (flag_format == 't') 
     {
@@ -495,10 +526,10 @@ dopage(int pageno)
     {
       if (! strcmp(outputfilename,"-")) {
         fout = stdout;
-#if defined(WIN32)
-        _setmode(_fileno(fout), _O_BINARY);
-#elif defined(__CYGWIN32__)
+#if defined(__CYGWIN32__)
         setmode(fileno(fout), O_BINARY);
+#elif defined(WIN32)
+        _setmode(_fileno(fout), _O_BINARY);
 #endif
       } else if (! (fout = fopen(outputfilename, "wb")))
         die(i18n("Cannot open output file '%s'."), outputfilename);
