@@ -51,8 +51,8 @@
 //C- | MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 //C- +------------------------------------------------------------------
 //
-// $Id: djvused.cpp,v 1.21 2004/04/13 13:54:57 leonb Exp $
-// $Name: release_3_5_14 $
+// $Id: djvused.cpp,v 1.24 2005/05/27 17:36:23 leonb Exp $
+// $Name: release_3_5_15 $
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -80,6 +80,7 @@
 #include "DataPool.h"
 #include "DjVuPort.h"
 #include "DjVuFile.h"
+#include "DjVmNav.h"
 
 static bool modified = false;
 static bool verbose = false;
@@ -918,13 +919,24 @@ print_meta(IFFByteStream &iff, ByteStream &out)
 void 
 command_print_meta(ParsingByteStream &)
 {
-  if (!g().file)
-    verror("you must first select a single page");
-  GP<ByteStream> out=ByteStream::create("w");
-  GP<ByteStream> anno = g().file->get_anno();
-  if (! (anno && anno->size())) return;
-  GP<IFFByteStream> iff=IFFByteStream::create(anno); 
-  print_meta(*iff,*out);
+  if (! g().file )
+    {
+      GP<DjVmDir::File> frec = g().doc->get_djvm_dir()->get_shared_anno_file();
+      if (frec)
+        {
+          vprint("print-meta: implicitely selecting shared annotations");
+          select_clear();
+          select_add(frec);
+        }
+    }
+  if ( g().file )
+    {
+      GP<ByteStream> out=ByteStream::create("w");
+      GP<ByteStream> anno = g().file->get_anno();
+      if (! (anno && anno->size())) return;
+      GP<IFFByteStream> iff=IFFByteStream::create(anno); 
+      print_meta(*iff,*out);
+    }
 }
 
 
@@ -995,8 +1007,6 @@ command_remove_meta(ParsingByteStream &)
 void
 command_set_meta(ParsingByteStream &pbs)
 {
-  if (!g().file)
-    verror("you must first select a single page");
   // get metadata
   GP<ByteStream> metastream = ByteStream::create();
   get_data_from_file("set-meta", pbs, *metastream);
@@ -1023,9 +1033,30 @@ command_set_meta(ParsingByteStream &pbs)
       if (key.length()>0 && val.length()>0)
         metadata[key] = val;
     }
+  // possibly select shared annotations.
+  if (! g().file)
+    {
+      GP<DjVmDir::File> frec = g().doc->get_djvm_dir()->get_shared_anno_file();
+      if (frec)
+        {
+          vprint("print-meta: implicitely selecting shared annotations.");
+        }
+      else if (metadata.size() > 0)
+        {
+          vprint("print-meta: implicitely creating and selecting shared annotations.");
+          g().doc->create_shared_anno_file();
+          frec = g().doc->get_djvm_dir()->get_shared_anno_file();
+        }
+      if (frec)
+        {
+          select_clear();
+          select_add(frec);
+        }
+    }
   // set metadata
-  if (modify_meta(g().file, &metadata))
-    vprint("set-meta: modified \"%s\"", (const char*)(GNativeString)g().fileid);
+  if (g().file && modify_meta(g().file, &metadata))
+    vprint("set-meta: modified \"%s\"", 
+           (const char*)(GNativeString)g().fileid);
 }
 
 struct  zone_names_struct
@@ -1445,6 +1476,113 @@ command_output_all(ParsingByteStream &)
     }
 }
 
+void
+print_outline_sub(const GP<DjVmNav> &nav, int &pos, int count, 
+                  const GP<ByteStream> &out, int indent)
+{
+  GUTF8String str;
+  GP<DjVmNav::DjVuBookMark> entry;
+  while (count > 0 && pos < nav->getBookMarkCount())
+    {
+      out->write("\n",1);
+      for (int i=0; i<indent; i++)
+        out->write(" ",1);
+      nav->getBookMark(entry, pos++);
+      out->write("(",2);
+      str = entry->displayname;
+      print_c_string(str, str.length(), *out);
+      out->write("\n ",2);
+      for (int i=0; i<indent; i++)
+        out->write(" ",1);
+      str = entry->url;
+      print_c_string(str, str.length(), *out);
+      print_outline_sub(nav, pos, entry->count, out, indent+1);
+      out->write(" )",2);
+      count--;
+    }
+}
+
+void
+command_print_outline(ParsingByteStream &pbs)
+{
+  GP<DjVmNav> nav = g().doc->get_djvm_nav();
+  if (nav)
+    {
+      int pos = 0;
+      int count = nav->getBookMarkCount();
+      if (count > 0)
+        {
+          const GP<ByteStream> out = ByteStream::create("w");
+          out->write("(bookmarks",10);
+          print_outline_sub(nav, pos, count, out, 1);
+          out->write(" )\n", 3);
+        }
+    }
+}
+
+void
+construct_outline_sub(ParsingByteStream &pbs, GP<DjVmNav> nav, int &count)
+{
+  int c;
+  GUTF8String name, url;
+  GP<DjVmNav::DjVuBookMark> mark;
+  if ((c = pbs.get_spaces(true)) != '\"')
+    verror("Syntax error in outline data: expecting name string.");    
+  pbs.unget(c);
+  name = pbs.get_utf8_token();
+  if ((c = pbs.get_spaces(true)) != '\"')
+    verror("Syntax error in outline data: expecting url string.");    
+  pbs.unget(c);
+  url = pbs.get_utf8_token();
+  mark = DjVmNav::DjVuBookMark::create(0, name, url);
+  nav->append(mark);
+  count += 1;
+  while ((c = pbs.get_spaces(true)) == '(')
+    construct_outline_sub(pbs, nav, mark->count);
+  if (c != ')')
+    verror("Syntax error in outline data: expecting ')'.");
+}
+
+GP<DjVmNav>
+construct_outline(ParsingByteStream &pbs)
+{
+  GP<DjVmNav> nav(DjVmNav::create());
+  int c = pbs.get_spaces(true);
+  int count = 0;
+  if (c == EOF)
+    return 0;
+  if (c!='(')
+    verror("Syntax error in outline data: expecting '(bookmarks ...'");
+  if (pbs.get_utf8_token()!="bookmarks")
+    verror("Syntax error in outline data: expecting 'bookmarks ...'");    
+  while ((c = pbs.get_spaces(true)) == '(')
+    construct_outline_sub(pbs, nav, count);
+  if (c != ')')
+    verror("Syntax error in outline data: expecting ')'.");
+  if (pbs.get_spaces(true) != EOF)
+    verror("Syntax error in outline data: garbage after last ')'");
+  if (nav->getBookMarkCount() < 1)
+    return 0;
+  if (!nav->isValidBookmark())
+    verror("Invalid outline data!");
+  return nav;
+}
+
+void
+command_set_outline(ParsingByteStream &pbs)
+{
+  const GP<ByteStream> outbs(ByteStream::create());
+  get_data_from_file("set-outline", pbs, *outbs);
+  outbs->seek(0);
+  GP<ParsingByteStream> outpbs(ParsingByteStream::create(outbs));
+  GP<DjVmNav> nav(construct_outline(*outpbs));
+  if (g().doc->get_djvm_nav() != nav)
+    {
+      g().doc->set_djvm_nav(nav);
+      modified = true;
+    }
+}
+
 static bool
 callback_thumbnails(int page_num, void *)
 {
@@ -1587,16 +1725,19 @@ command_help(void)
           " . print-meta             -- prints file metadatas (a subset of the annotations\n"
           "   print-txt              -- prints hidden text using a lisp syntax\n"
           "   print-pure-txt         -- print hidden text without coordinates\n"
+          " _ print-outline          -- print outline (bookmarks)\n"
           "   output-ant             -- dumps ant as a valid cmdfile\n"
           "   output-txt             -- dumps text as a valid cmdfile\n"
           "   output-all             -- dumps ant and text as a valid cmdfile\n"
           " . set-ant [<antfile>]    -- copies <antfile> into the annotation chunk\n"
           " . set-meta [<metafile>]  -- copies <metafile> into the metadata part of the annotations\n"
           " . set-txt [<txtfile>]    -- copies <txtfile> into the hidden text chunk\n"
+          " _ set-outline [<bmfile>] -- sets outline (bootmarks)\n"
           " _ set-thumbnails [<sz>]  -- generates all thumbnails with given size\n"
           "   remove-ant             -- removes annotations\n"
           "   remove-meta            -- removes metadatas without changing other annotations\n"
           "   remove-txt             -- removes hidden text\n"
+          " _ remove-outline         -- removes outline (bookmarks)\n"
           " _ remove-thumbnails      -- removes all thumbnails\n"
           " . save-page <name>       -- saves selected page/file as is\n"
           " . save-page-with <name>  -- saves selected page/file, inserting all included files\n"
@@ -1647,12 +1788,14 @@ static GMap<GUTF8String,CommandFunc> &command_map() {
     xcommand_map["print-meta"] = command_print_meta;
     xcommand_map["print-txt"] = command_print_txt;
     xcommand_map["print-pure-txt"] = command_print_pure_txt;
+    xcommand_map["print-outline"] = command_print_outline;
     xcommand_map["output-ant"] = command_output_ant;
     xcommand_map["output-txt"] = command_output_txt;
     xcommand_map["output-all"] = command_output_all;
     xcommand_map["set-ant"] = command_set_ant;
     xcommand_map["set-meta"] = command_set_meta;
     xcommand_map["set-txt"] = command_set_txt;
+    xcommand_map["set-outline"] = command_set_outline;
     xcommand_map["set-thumbnails"] = command_set_thumbnails;
     xcommand_map["remove-ant"] = command_remove_ant;
     xcommand_map["remove-meta"] = command_remove_meta;
